﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using Settings;
using UnityEngine.SceneManagement;

public class GameManager : MonoBehaviour {

    /// <summary>
    /// The player prefab to spawn.
    /// </summary>
    public GameObject PlayerPrefab;

    /// <summary>
    /// The players
    /// </summary>
    public Player[] Players;

    /// <summary>
    /// The settings of the game
    /// </summary>
    PlayerSettings settings;

    // If the player is currently spawning a player or not.
    private bool isSpawingPlayer = false;

    /// <summary>
    /// The vibration manager for all controllers
    /// </summary>
    VibrationManager vibrator;

    ScoreContainerController scoreGUI;

    /// <summary>
    /// The current game mode
    /// </summary>
    GameMode gameMode;

    void Awake()
    {
        gameMode = new RoundsMode();
    }

    // Use this for initialization
    void Start()
    {
        // Turn off all controllers
        vibrator = GetComponent<VibrationManager>();
        vibrator.StopAllVibration();

        // Create the correct amount of players
        settings = FindObjectOfType<PlayerSettings>();
        if (!settings)
        {
            settings = new GameObject("PlayerSettings", typeof(PlayerSettings)).GetComponent<PlayerSettings>();
            settings.inputs = new List<InputType>() { InputType.Keyboard, InputType.Mouse };
        }

        List<int> controllerIDs = new List<int>(settings.addedControllerIDs);
        Players = new Player[settings.inputs.Count];

        // Reset static playercount variable to 1 (we have at least one player)
        PlayerManager.PlayerCount = 1;

        // Initialize players
        int playerIndex = 0;
        GameObject go;
        foreach (InputType type in settings.inputs)
        {
            // Instantiate new player
            go = Instantiate(PlayerPrefab, GetRandomSpawnPosition(), Quaternion.identity) as GameObject;
            go.transform.rotation = Quaternion.Euler(0, 0, Random.Range(0, 360));

            // Create new player object
            Players[playerIndex] = new Player(go.GetComponent<PlayerManager>(), type);

            // Set player alive
            Players[playerIndex].Alive = true;

            switch (Players[playerIndex].Input)
            {
                case InputType.Keyboard:
                    go.AddComponent<KeyboardInput>();
                    break;
                case InputType.Mouse:
                    go.AddComponent<MouseInput>();
                    break;
                case InputType.Controller:
                    ControllerInput input = go.AddComponent<ControllerInput>();
                    input.ControllerID = controllerIDs[0] + 1;
                    controllerIDs.RemoveAt(0);
                    vibrator.AddPlayerManager(Players[playerIndex].Manager);
                    break;
                default:
                    return;
            }

            Players[playerIndex].Manager.DieEvent.AddListener(OnPlayerDie);
            playerIndex++;
        }

        Camera.main.GetComponentInChildren<CountdownManager>().StartCountdown();

        // Init the gui
        scoreGUI = FindObjectOfType<ScoreContainerController>();
        scoreGUI.InitializeScoreController(Players);
    }
	
	// Update is called once per frame
	void Update () {
	
	}

    public void UnloadScene(string level)
    {
        Destroy(settings.gameObject);
        SceneManager.LoadScene(level);     // Go to menu
    }

    /// <summary>
    /// Inits the game for a new 'round'.
    /// </summary>
    void InitGame()
    {
        // Destroy old line managers 
        PlayerLineManager[] lineManagers = FindObjectsOfType<PlayerLineManager>();
        foreach (PlayerLineManager lm in lineManagers)
        {
            Destroy(lm.gameObject);
        }

        // Spawn players
        foreach (Player player in Players)
        {
            StartCoroutine(SpawnPlayer(player));
        }
        // Do countdown
        Camera.main.GetComponentInChildren<CountdownManager>().StartCountdown();
    }

    /// <summary>
    /// Called when a player has died
    /// </summary>
    /// <param name="PlayerID">The player who died</param>
    public void OnPlayerDie(int PlayerID)
    {
        Player player = GetPlayerByID(PlayerID);
        player.Alive = false;

        Player winningPlayer;
        if (gameMode.IsGameWon(Players, out winningPlayer))
        {
            winningPlayer.Score++;              // Add score to player
            scoreGUI.SetScore(winningPlayer);   // Update UI
            if (winningPlayer.Score < 5)
            {
                Camera.main.GetComponentInChildren<CountdownManager>().ShowRoundWinner(winningPlayer, () => {
                    InitGame();                         // Restart game
                });
            }
            else
            {
                Camera.main.GetComponentInChildren<CountdownManager>().ShowGameWinner(winningPlayer, () => {
                    UnloadScene("menu");
                });
            }
        }
    }


    /// <summary>
    /// Returns a player by his ID number
    /// </summary>
    /// <param name="playerID">The ID of the player to look for.</param>
    /// <returns>Returns the player with the given ID if he exists. Else returns null.</returns>
    private Player GetPlayerByID(int playerID)
    {
        Player result = null;
        foreach(Player player in Players)
        {
            if (player.ID == playerID)
            {
                result = player;
            }
        }
        return result;
    }

    /// <summary>
    /// Creates a new player
    /// </summary>
    /// <param name="PlayerID"></param>
    /// <returns></returns>
    IEnumerator RespawnPlayer(int PlayerID)
    {
        while(isSpawingPlayer)                                  // Never spawn two players at the same time
        {
            yield return new WaitForEndOfFrame();
        }
        isSpawingPlayer = true;
        
        Player player = GetPlayerByID(PlayerID);                // Find gameobject with the correct id
        GameObject go = player.Manager.gameObject;              // Get the gameobject of the player

        go.transform.position = GetRandomSpawnPosition();       // Jump player to new spawn position
        go.transform.rotation = Quaternion.Euler(0, 0,
                                    Random.Range(0, 360));      // Add random angle

        go.GetComponent<PlayerMotor>().canMove = false;         // Player can't move untill countdown.
        go.SetActive(true);                                     // Player should be active
        player.Alive = true;                                    // Set player alive

        yield return StartCoroutine(Players[PlayerID - 1].Manager.GetComponentInChildren<CountdownManager>().DoCountdown());

        isSpawingPlayer = false;                                // Let another player spawn
    }

    /// <summary>
    /// Creates a new player
    /// </summary>
    /// <param name="PlayerID"></param>
    /// <returns></returns>
    IEnumerator SpawnPlayer(Player player)
    {
        while (isSpawingPlayer)                                  // Never spawn two players at the same time
        {
            yield return new WaitForEndOfFrame();
        }
        isSpawingPlayer = true;

        GameObject go = player.Manager.gameObject;              // Get the gameobject of the player

        go.transform.position = GetRandomSpawnPosition();       // Jump player to new spawn position
        go.transform.rotation = Quaternion.Euler(0, 0, 
                                    Random.Range(0, 360));      // Add random angle

        PlayerMotor pm = go.GetComponent<PlayerMotor>();
        pm.CreateNewLine();                                     // Create a new line
        pm.canMove = false;                                     // Player can't move untill countdown.

        go.SetActive(true);                                     // Player should be active
        player.Alive = true;                                    // Set player alive
        isSpawingPlayer = false;                                // Let another player spawn
    }

    /// <summary>
    /// Calculates a new random position within the screen borders
    /// </summary>
    /// <returns>The random spawn position</returns>
    Vector2 GetRandomSpawnPosition()
    {
        Vector2 result = Vector2.zero;
        float padding = 0.2f;

        int tries = 0;
        do
        {
            float x = Random.Range(padding, (1.0f - padding));    // Normalized random positions
            float y = Random.Range(padding, (1.0f - padding));
            result = new Vector2(x, y);
            result = Camera.main.ViewportToWorldPoint(new Vector3(result.x, result.y, -transform.position.z));  // Get world position based on normalized position
            tries++;
        }
        while (IsPositionNearPlayer(Players, result) && tries < 100);

        return result;
    }

    /// <summary>
    /// Checks if the position is near one of the players
    /// </summary>
    /// <param name="players">The other players</param>
    /// <param name="position">The positon to check</param>
    /// <returns>Weighter the position is near any of the players</returns>
    bool IsPositionNearPlayer(Player[] players, Vector3 position)
    {
        foreach (Player p in players)
        {
            if (p != null && Vector3.Distance(position, p.Manager.gameObject.transform.position) < 5)
            {
                return true;
            }
        }
        return false;
    }
}
