﻿using UnityEngine;
using System.Collections.Generic;

[RequireComponent(typeof(MeshRenderer))]
[RequireComponent(typeof(MeshFilter))]
[RequireComponent(typeof(MeshFilter))]
public class PlayerLineManager : MonoBehaviour {

    /// <summary>
    /// The color that the line manager needs to draw
    /// </summary>
    public Color color;

    // Keeps track of how many point updates there have been
    private int count = 0;

    /// <summary>
    /// Once in the how many point updates it should create a new collider
    /// </summary>
    public int colliderFrequency = 3;

    /// <summary>
    /// Used to set the correct color for the liner renderer
    /// </summary>
    private new MeshRenderer renderer;

    /// <summary>
    /// Gets the mesh fitler for setting and creating the line mesh
    /// </summary>
    private MeshFilter filter;

    /// <summary>
    /// How often should we create a gap?
    /// </summary>
    [Range(0.0f, 1.0f)]
    public float gapChance = 0.2f;
    /// <summary>
    /// The minimum gap width
    /// </summary>
    public float minGapWidth = 0.5f;
    /// <summary>
    /// The maximum gap width
    /// </summary>
    public float maxGapWidth = 2f;
    /// <summary>
    /// Are we currently making a gap?
    /// </summary>
    private bool gap = false;
    /// <summary>
    /// If we had a gap last frame
    /// </summary>
    private int pointsSinceGap = int.MaxValue;
    /// <summary>
    /// Amount of steps to skip before drawing the line again.
    /// </summary>
    private int gapIterations = 0;


    /// <summary>
    /// The mesh of the line manager.
    /// </summary>
    private Mesh mesh;

    /// <summary>
    /// Recorded points from the player
    /// </summary>
    private List<Vector2> points = new List<Vector2>();

    /// <summary>
    /// The height that the box colliders have.
    /// </summary>
    float colliderHeight = 0.05f;
    /// <summary>
    /// All collider having game objects
    /// </summary>
    private List<GameObject> colliders = new List<GameObject>();

    // Use this for initialization
    void Start ()
    {
        filter = GetComponent<MeshFilter>();
        GetComponent<MeshRenderer>().material.color = color;
    }
	
    /// <summary>
    /// Sends a point to the line renderer so it can render a new line segment
    /// </summary>
    /// <param name="point">Next point the line goes to</param>
    /// <param name="width">The width of this line point</param>
    public void CreatePoint(Transform point, float width)
    {
        if (filter == null)
        {
            return;
        }

        mesh = filter.mesh;

        // Check if we need to create a mesh.
        if (mesh == null)
        {
            mesh = new Mesh();
        }

        // Get the vertices and triangles from the mesh
        Vector3[] vertices = mesh.vertices;
        int[] triangles = mesh.triangles;

        if (gap && gapIterations > 0)
        {
            gapIterations--;
            return; 
        }
        else if (gap && gapIterations <= 0)
        {
            gap = false;
            vertices = AddToArray(vertices, point.position - point.right * width);
            vertices = AddToArray(vertices, point.position + point.right * width);
        }

        // Add three vetices 
        // *----x----*
        vertices = AddToArray(vertices, point.position - point.right * width);
        vertices = AddToArray(vertices, point.position + point.right * width);

        // Add to the points list
        points.Add(point.position);
        pointsSinceGap++;
        
        // For every colliderFrequency index we create a cericle collider 2d.
        if (!gap && count++ % colliderFrequency == 0 && points.Count > 2)
        {
            GameObject go = new GameObject("Box Collider 2D");
            go.tag = transform.tag;
            // Add box collider
            BoxCollider2D bo = go.AddComponent<BoxCollider2D>();
            bo.size = new Vector2(width * 2 - 0.05f, colliderHeight);
            bo.offset = new Vector2(0, -bo.size.y * 0.5f);
            if (pointsSinceGap == 1)
            {
                go.transform.position = point.position + point.up * colliderHeight;
                go.transform.rotation = point.rotation;
                bo.enabled = false;
            }
            if (pointsSinceGap == 3)
            {
                colliders[colliders.Count - 1].GetComponent<BoxCollider2D>().enabled = true;
            }
            else
            {
                go.transform.position = point.position;
                go.transform.rotation = point.rotation;
            }
            go.transform.SetParent(transform);

            colliders.Add(go);
        }

        // Here connect the vertices to the path with triangels.
        if (vertices.Length >= 6)
        {
            triangles = AddTriangle(triangles,
                vertices.Length - 4,
                vertices.Length - 3,
                vertices.Length - 2, 
                vertices.Length - 1);
        }

        // Update and optimize mesh
        mesh.SetVertices(new List<Vector3>(vertices));
        mesh.SetTriangles(triangles, 0);
        mesh.Optimize();

        // Update the real mesh
        filter.mesh = mesh;

        // Decide if we should make a gap next time
        float chance = Random.Range(0.0f, 1.0f);
        if (chance <= gapChance)
        {
            // Make gap
            gap = true;
            pointsSinceGap = 0;
            gapIterations = (int)maxGapWidth;
            vertices = AddToArray(vertices, point.position - point.right * width);
            vertices = AddToArray(vertices, point.position + point.right * width);
        }
    }

    /// <summary>
    /// Adds an element to the array
    /// </summary>
    /// <typeparam name="T">The type of the array</typeparam>
    /// <param name="array">The array to add element to</param>
    /// <param name="element">The element to add</param>
    /// <returns></returns>
    public T[] AddToArray<T>(T[] array, T element)
    {
        T[] newArray = new T[array.Length + 1];
        array.CopyTo(newArray, 0);
        newArray[array.Length] = element;
        return newArray;
    }

    /// <summary>
    /// Make triangle from four vertices.
    /// The vertex layout is used like this:
    /// 2 --- 3
    /// | \   |
    /// |   \ |
    /// 0 --- 1
    /// </summary>
    /// <param name="tries">The </param>
    /// <param name="p0"></param>
    /// <param name="p1"></param>
    /// <param name="p2"></param>
    /// <param name="p3"></param>
    private int[] AddTriangle(int[] tries, int p0, int p1, int p2, int p3)
    {
        // Make the two triangles
        tries = AddToArray(tries, p1);  // Tri 1
        tries = AddToArray(tries, p0);
        tries = AddToArray(tries, p2);

        tries = AddToArray(tries, p1);  // Tri 2
        tries = AddToArray(tries, p2);
        tries = AddToArray(tries, p3);

        return tries;
    }

    void OnDrawGizmos()
    {
        Gizmos.color = Color.magenta;
        foreach (Vector2 p in points)
        {
            Gizmos.DrawWireSphere(p, 0.01f);
        }
    }
}
