﻿using UnityEngine;
using UnityEngine.Events;
using System.Collections;
using System;

[System.Serializable]
public class OnPlayerDieEvent : UnityEvent<int> { }

public class PlayerManager : MonoBehaviour {

    /// <summary>
    /// The ID of the player, 
    /// </summary>
    public int ID = -1;
    /// <summary>
    /// The total player count should always be bigger then 1
    /// </summary>
    public static int PlayerCount = 1;

    /// <summary>
    /// The ripple effect that is shown when dying
    /// </summary>
    public GameObject RippleEffect;

    /// <summary>
    /// This is the color of the player we are
    /// </summary>
    public Color color
    {
        get { return playerColors[ID - 1]; }
    }
    /// <summary>
    /// Color coded player colors
    /// </summary>
    public static Color[] playerColors = new Color[] {
        Color.magenta,
        Color.yellow,
        Color.cyan,
        Color.red,
        Color.green,
        Color.blue
    };

    /// <summary>
    /// Event called when the player is about to die
    /// </summary>
    public OnPlayerDieEvent DieEvent;

    // Use this for initialization
    void Awake ()
    {
        // Set player id if it is not set correctly
        if (ID == -1)
        {
            ID = PlayerCount;
            PlayerCount++;
        }

        // Setup event
        if (DieEvent == null)
        {
            DieEvent = new OnPlayerDieEvent();
            DieEvent.AddListener(KillPlayer);
        }
    }

    // Update is called once per frame
    void Update()
    {
        int padding = 20;
        float paddingFraqX = (float)padding / (float)Camera.main.pixelWidth;
        float paddingFraqY = (float)padding / (float)Camera.main.pixelHeight;
        // Check screen bounds
        Vector2 screenPos = Camera.main.WorldToViewportPoint(transform.position);
        if (screenPos.x < paddingFraqX || screenPos.x > 1.0f - paddingFraqX || 
            screenPos.y < paddingFraqY || screenPos.y > 1.0f - paddingFraqY)
        {
            KillPlayer(ID);
            DieEvent.Invoke(ID);
        }
    }

    /// <summary>
    /// Gets called when we collide with a line
    /// </summary>
    /// <param name="other"></param>
    void OnCollisionEnter2D(Collision2D other)
    {
        switch (other.gameObject.tag)
        {
            case "Player":
                if (other.gameObject != gameObject)
                {
                    KillPlayer(ID);
                    DieEvent.Invoke(ID);
                }
                break;
            case "Line":
                KillPlayer(ID);
                DieEvent.Invoke(ID);
                break;
            default:
                break;
        }
    }

    void OnTriggerEnter2D(Collider2D other)
    {
        print(other.gameObject.name);
        switch (other.gameObject.tag)
        {
            case "Player":
                if (other.gameObject != gameObject)
                {
                    KillPlayer(ID);
                    DieEvent.Invoke(ID);
                }   
                break;
            case "Line":
                KillPlayer(ID);
                DieEvent.Invoke(ID);
                break;
            default:
                break;
        }
    }

    /// <summary>
    /// Destroys the player line and induces screenshake. Activates the player die event.
    /// </summary>
    /// <param name="ID">The id value of the player who calls the event.</param>
    public void KillPlayer(int ID)
    {
        ScreenShake.Instance.Shake(0.2f, 0.3f, 0.05f);

        if (RippleEffect)
        {
            Instantiate(RippleEffect, transform.position, Quaternion.identity);
        }

        PlayerCount--;
        Destroy(GetComponent<PlayerMotor>().lineManager.gameObject);
        gameObject.SetActive(false);
    }
}
