﻿using UnityEngine;

/// <summary>
/// The class that moves the player. Takes input from an input source
/// </summary>
public class PlayerMotor : MonoBehaviour
{
    /// <summary>
    /// The speed which we travel at
    /// </summary>
    public float speed;
    /// <summary>
    /// The maximum radius we can travel in
    /// </summary>
    public float maxDegreesDelta;

    /// <summary>
    /// If the player can move
    /// </summary>
    public bool canMove = false;

    /// <summary>
    /// A prefab of the direction indicator
    /// </summary>
    public GameObject directionIndicator;
    // The current indicator for the direction
    private GameObject _dirIndicator;
    // The distance between the direction indicator and the player object
    public float directionIndicatorDistance;

    /// <summary>
    /// A prefab for a particular line render
    /// </summary>
    public GameObject lineManagerPrefab;
    /// <summary>
    /// The line manager for this particular player
    /// </summary>
    [HideInInspector]
    public PlayerLineManager lineManager;
    /// <summary>
    /// How many times per second we update our position to the line renderer
    /// </summary>
    public int lineUpdateFrequency = 10;
    /// <summary>
    /// Current width of the line
    /// </summary>
    public float lineWidth;

    /// <summary>
    /// The player manager
    /// </summary>
    public PlayerManager playerManager;

    // Use this for initialization
    void Start()
    {
        // Get player info
        playerManager = GetComponent<PlayerManager>();

        // Create a line manager
        CreateNewLine();

        // Create new indic
        if (_dirIndicator == null)
        {
            _dirIndicator = Instantiate(directionIndicator, transform.position, transform.rotation) as GameObject;
            _dirIndicator.transform.SetParent(transform);
        }

        // Invoke the update line each 1/frequency seconds
        InvokeRepeating("UpdateLine", 1.0f / lineUpdateFrequency, 1.0f / lineUpdateFrequency);
    }

    void OnEnable()
    {
        if (playerManager.ID > 0)
        {
            CreateNewLine();
            transform.rotation = Quaternion.identity;
        }
    }

    Vector3 direction;

    // Update is called once per frame
    public void UpdatePosition(float input)
    {
        // Get Input
        if (Mathf.Abs(input) < 0.1f)
        {
            input = 0;
        }

        // Set graphics
        if (_dirIndicator)
        {
            _dirIndicator.transform.position = transform.position + transform.right * input * directionIndicatorDistance;
        }
        
        if (canMove)
        {
            transform.Rotate(Vector3.forward, -input * maxDegreesDelta * Time.deltaTime);
            transform.position += transform.up * speed * Time.deltaTime;
        }
    }

    /// <summary>
    /// Updates the line manager by supplying new point
    /// </summary>
    void UpdateLine()
    {
        if (canMove)
        {
            lineManager.CreatePoint(transform, lineWidth);
        }
    }

    /// <summary>
    /// Creates and binds a reference to a new line.
    /// </summary>
    public void CreateNewLine()
    {
        GameObject go = Instantiate(lineManagerPrefab) as GameObject;
        lineManager = go.GetComponent<PlayerLineManager>();
        lineManager.color = playerManager.color;
    }
}
