﻿using UnityEngine;
using System.Collections;

public class ScreenShake : MonoBehaviour
{

    /// <summary>
    /// Singleton instance of screenshake
    /// </summary>
    public static ScreenShake Instance;

    /// <summary>
    /// Default tooltip time
    /// </summary>
    [Tooltip("The default duration of the screenshake.")]
    public float defaultDuration = 0.1f;

    /// <summary>
    /// The default amplitude that it travels
    /// </summary>
    [Tooltip("The default amplitude the camera travels")]
    public float defaultAmplitude = 0.5f;

    /// <summary>
    /// The position the camera started at
    /// </summary>
    private Vector3 startPosition;

    public RGBEffect rgbEffect;

    // Use this for initialization
    void Start()
    {
        // Singleton pattern
        if (Instance == null)
            Instance = this;
        else if (Instance != this)
        {
            Debug.LogWarning("Second screen shake created, deleting this script!");
            Destroy(this);
        }

        // Set start position
        startPosition = transform.position;
    }

    /// <summary>
    /// Shakes the screen
    /// </summary>
    public void Shake()
    {
        Shake(defaultDuration, defaultAmplitude);
    }

    /// <summary>
    /// Shakes the screen for a certain time
    /// </summary>
    /// <param name="duration"></param>
    public void ShakeDuration(float duration)
    {
        Shake(duration, defaultAmplitude);
    }

    /// <summary>
    /// Shakes the screen with a certain amplitude
    /// </summary>
    /// <param name="amplitude">The amplitdue of the shaking</param>
    public void ShakeAmplitude(float amplitude)
    {
        Shake(defaultDuration, amplitude);
    }

    /// <summary>
    /// Shakes the screen with a certain duration and amplitude
    /// </summary>
    /// <param name="duration">How long to shake</param>
    /// <param name="amplitude">The amplitude of the shaking</param>
    public void Shake(float duration, float amplitude)
    {
        StartCoroutine(ShakeScreen(defaultDuration, amplitude));
    }

    /// <summary>
    /// Shakes the screen towards a direction
    /// </summary>
    /// <param name="direction"></param>
    public void Shake(Vector2 direction)
    {
        Shake(defaultDuration, direction);
    }

    /// <summary>
    /// Shakes the screen towards a dirction and with a certain duration
    /// </summary>
    /// <param name="duration">Duration of the shaking</param>
    /// <param name="direction"></param>
    public void Shake(float duration, Vector2 direction)
    {
        StartCoroutine(BounceScreen(direction, duration));
    }

    public void Shake(float amplitude, float frequency, float duration)
    {
        StartCoroutine(ShakeScreen(duration, amplitude));
    }

    /// <summary>
    /// Shakes the screen for a time
    /// </summary>
    /// <param name="duration">How long to shake</param>
    /// <returns></returns>
    IEnumerator ShakeScreen(float duration, float amplitude)
    {
        float startTime = Time.unscaledTime;
        while (Time.unscaledTime - startTime < duration)
        {
            Vector2 rndm = Random.insideUnitCircle;
            transform.position = startPosition + (Vector3)rndm * amplitude;
            if (rgbEffect)
            {
                rgbEffect.redShift = rndm * 0.02f * amplitude;
                rgbEffect.greenShift = new Vector2(-rndm.x, rndm.y) * 0.02f * amplitude;
                rgbEffect.blueShift = new Vector2(-rndm.x, -rndm.y) * 0.02f * amplitude;

            }
            yield return new WaitForEndOfFrame();
        }
        // Reset to start position always
        transform.position = startPosition;
        if (rgbEffect)
        {
            transform.position = startPosition;
            rgbEffect.redShift = Vector2.zero;
            rgbEffect.greenShift = Vector2.zero;
            rgbEffect.blueShift = Vector2.zero;
        }
    }

    /// <summary>
    /// Bounces the screen to a direction and back.
    /// </summary>
    /// <param name="direction">Where to bounce towards (local space)</param>
    /// <param name="duration">How long to stay there</param>
    /// <returns></returns>
    IEnumerator BounceScreen(Vector2 direction, float duration)
    {
        // Jump to direction
        transform.position += (Vector3)direction;
        if (rgbEffect)
        {
            rgbEffect.redShift = new Vector2(direction.x, direction.y).normalized * 0.01f;
            rgbEffect.greenShift = new Vector2(-direction.x, direction.y).normalized * 0.01f;
            rgbEffect.blueShift = new Vector2(direction.x, -direction.y).normalized * 0.01f;

        }
        // Wait for a bit
        yield return new WaitForSeconds(duration);
        // Reset to start position always
        if (rgbEffect)
        {
            transform.position = startPosition;
            rgbEffect.redShift = Vector2.zero;
            rgbEffect.greenShift = Vector2.zero;
            rgbEffect.blueShift = Vector2.zero;
        }
    }
}